#!/usr/bin/env python
# Copyright 2008-2009 David D Lowe
# Released under the GPLv3
"""Contains the Shell class and its subclasses"""

from __future__ import division
import const
from const import DEBUG, MY_CACHE_HOME, MY_DATA_HOME, PIGMENT_TYPES
import pygtk
pygtk.require("2.0")
import gtk
import pexpect
import os, sys
import time
from gettext import gettext as _
import subprocess

DEBUG_SHELL = True
if not DEBUG:
    DEBUG_SHELL = False

def bash_prepare(string): 
    """Prepares a string for bash by surrounding it with single quotes.
    >>> shell.bash_prepare("two'words")
    "'two'\\''words'"
    """ 
    return "'" + string.replace("'", r"'\''") + "'"

class Shell():
    """An object which can run commands
    This is an "abstract" class, only its subclasses should be used"""
    def __init__(self):
        self.ready = False
    
    def prepare(self):
        """Prepares the shell, may ask user for password if necessary"""
        return self.ready
    
    def do(self, command, intrShell=False, cwd=None):
        """Runs the command in this shell.
        If intrShell is True, allows the shell to interpret special characters
        such as *
        cwd: the current directory for the command"""
        pass
    
    def set_out_file(self, outfile):
        """Sets the file where stdout output is put"""
        pass
    
    def exit(self):
        """Exits the shell"""
        self.ready = False

class BashRootShell(Shell):
    """The Shell object which has access to a bash terminal run as root
    It uses pexpect, and it runs sudo bash -i"""
    def __init__(self, prompt="epidermisbash:"):
        Shell.__init__(self)
        self.prompt = prompt
        self.logfile = os.tmpfile()
        self.gksubash, self.gksued = None, None
        self.spawn = None
        self.ready = False
        import epidermis
        for pp in ("/usr/share/epidermis/data/gksubash.sh", os.path.join(epidermis.determine_path(), "data", "gksubash.sh")):
            if os.path.exists(pp):
                self.gksubash = pp
        for pp in ("/usr/share/epidermis/data/gksued.sh", os.path.join(epidermis.determine_path(), "data", "gksued.sh")):
            if os.path.exists(pp):
                self.gksued = pp
        if self.gksubash is None:
            raise(Exception("Cannot find gksubash.sh"))
        if self.gksued is None:
            raise(Exception("Cannot find gksued.sh"))
        customm = False
        for pp in self.gksubash, self.gksued:
            if pp[:21] != "/usr/share/epidermis/":
                customm = True
                break
        if customm: print "Using custom gksubash and gksued"
        
    def prepare(self):
        """Initializes the bash shell as root, will ask user for 
        password graphically if necessary
        Will raise a ShellNotReadyException if it fails to prepare itself
        # Unfortunately, due to bug #244930, gksu cannot run bash -i correctly
        # This workaround uses sudo, but still asks for the password
        # graphically, if required
        # See bug: https://bugs.launchpad.net/ubuntu/+source/gksu/+bug/244930"""
        self.spawn = pexpect.spawn("sh", [self.gksubash, self.gksued])
        self.spawn.logfile = self.logfile
        res = self.spawn.expect(["EPIDERMIS_PASS", "EPIDERMIS_SUDO_RUN_SUCCESSFULLY", \
            "EPIDERMIS_SUDO_FINISHED"])
        self.ready = True
        while res != 1: # sudo didn't run successfully
            if res == 2:
                # sudo exited unexpectedly
                self.ready = False
                raise(ShellNotReadyException())
                break
            # I need a password, use gksu gui to get one
            pop = os.popen("gksu -p --message " + bash_prepare(_("Enter password to allow Epidermis to access root privileges")))
            password = "IMPOSSIBLE_PASSWORD_ITHINK3457865"
            for line in pop.readlines():
                if "sn_launcher_context_complete" in line:
                    continue
                if len(line) == 0:
                    continue
                password = line[:-1]
            if password == "IMPOSSIBLE_PASSWORD_ITHINK3457865" or password == "":
                retry = self.yes_no_box(_("Could not retrieve password, try again?"))
                if not retry:
                    self.ready = False
                    raise(ShellNotReadyException())
                    break
                else:
                    continue
            # try this password, disable logfile to avoid printing password
            oldlogfile = self.spawn.logfile
            self.spawn.logfile = os.tmpfile()
            self.spawn.sendline(password)
            self.spawn.logfile = oldlogfile
            # recheck to see if successful
            res = self.spawn.expect(["EPIDERMIS_PASS", "EPIDERMIS_SUDO_RUN_SUCCESSFULLY", \
                "EPIDERMIS_SUDO_FINISHED"])
        
        self.spawn.readline()
        
        if not self.ready:
            raise(ShellNotReadyException())
        else:
            if hasattr(self, "logfile"):
                self.spawn.logfile = self.logfile
            ##self.spawn.sendline(" PS1=" + self.prompt)
            ##self.spawn.readline()
            ##self.spawn.expect(self.prompt)
            self.do(" PS1=" + self.prompt)
            self.do(" HISTCONTROL=\"ignoreboth\"")
            return True
    
    def set_logfile(self, logfile):
        """sets where the bash should print to, by default, no where
        logfile: file like object"""
        if not self.spawn is None:
            Sfile = logfile
        self.logfile = logfile
    
    def set_out_file(self, outfile):
        """This has same effect as set_logfile, although not quite correct"""
        self.set_logfile(outfile)
    
    def do(self, command, intrShell=False, cwd=None, expectStr=None, expectNum=1 ):
        """Runs the command
        command: the command string as bash would expect, or a list of command
        arguments
        expectStr: the string to stop execution at, if not specified, Shell will
        stop at bash prompt
        cwd: the current directory for the command run
        expectNum: if expectStr is specified, this number indicates the number
        of times do must expect expectStr
        intrShell: if False, do will automatically prepare bash on each of the
        command's list's items. Currently has no effect if command is str"""
        
        if not self.ready:
            raise(ShellNotReadyException())
            
        cmd = ""
        if isinstance(command, list):
            for item in command:
                if intrShell:
                    cmd = cmd + item.replace(" ", r"\ ") + " "
                else:
                    cmd = cmd + bash_prepare(item) + " "
            cmd = cmd[:-1]
        elif isinstance(command, str):
            cmd = command
        else:
            raise(Exception("command must be list or string"))
        
        if DEBUG_SHELL:
            print "debug+++root do: " + cmd
        if not cwd is None:
            self.spawn.sendline(" cd " + bash_prepare(cwd))
            self.spawn.readline()
            self.spawn.expect(self.prompt)
        self.spawn.sendline(" " + cmd) # space is necessary so that the command is not 
                                       # recorded in history
        self.spawn.readline() # reads the line in which the command was entered
        if expectStr is None:
            self.spawn.expect(self.prompt)
        else:
            coun = 0
            while coun < expectNum:
                self.spawn.expect(expectStr)
                coun += 1
            
    
    def yes_no_box(self, message):
        dd = gtk.MessageDialog(type=gtk.MESSAGE_ERROR, buttons=gtk.BUTTONS_YES_NO, \
            message_format=message)
        res = dd.run()
        dd.hide()
        dd.destroy()
        if res == gtk.RESPONSE_YES:
            return True
        else:
            return False
    
    def exit(self):
        """exits the bash shell, terminates it if necessary"""
        self.do("exit", expectStr="exit")
        
        if self.spawn.isalive():
            time.sleep(0.3)
            if self.spawn.isalive():
                self.spawn.terminate()
        self.ready = False

class SubprocessShell(Shell):
    """An object which can run commands using subprocess.call as current user"""
    def __init__(self):
        self.ready = True
        self.outFile = sys.stdout
        self.errFile = sys.stderr
    
    def prepare(self):
        """Prepares the shell"""
        self.ready = True
        self.outFile = sys.stdout
        return self.ready
    
    def do(self, command, intrShell=False, cwd=None):
        """Runs the command.
        command: can only be a list
        If intrShell is True, allows the shell to interpret special characters
        such as *
        cwd: current directory"""
        if not isinstance(command, list):
            raise(Exception("Not list"))
        global DEBUG_SHELL
        if DEBUG_SHELL:
            cmdStr = ""
            for it in command:
                cmdStr = cmdStr + it + " "
            if len(cmdStr) > 0:
                cmdStr = cmdStr[:-1]
            print "debug+++subprocess do: " + cmdStr
        subprocess.call(command, shell=intrShell, cwd=cwd, stdout=self.outFile, \
            stderr=self.errFile)
        
    def set_out_file(self, outfile):
        self.outFile = outfile
    
    def set_err_file(self, errfile):
        self.errFile = errfile
    
    def exit(self):
        """Exits the shell, this shell is always ready so this method has no
        function"""
        self.ready = True

class ShellNotReadyException(Exception):
    def __str__(self):
        return "Shell has not been prepared"

def main():
    print "don't run"

def testshell():
    """A function that tries some of BashRootShell's functionality"""
    shell = BashRootShell()
    print "hi"
    shell.set_logfile(sys.stdout)
    shell.prepare()
    print "done"
    if shell.ready == False:
        print "not ready, user cancelled?"
        return

    shell.do("ls")
    shell.do("whoami")
    
    shell.exit()

def determine_path ():
    """Borrowed from wxglade.py"""
    try:
        root = __file__
        if os.path.islink (root):
            root = os.path.realpath (root)
        return os.path.dirname (os.path.abspath (root))
    except Exception, ee:
        print >> sys.stderr, "I'm sorry, but something is wrong."
        print >> sys.stderr, "There is no __file__ variable. Please contact the author."
        sys.exit (1)
        raise(ee)


if __name__ == "__main__":
    main()
